/* PicasaWeb Downloader v1.2
 * Tomer Gabel, 2007-2008
 * http://www.tomergabel.com
 * 
 * Hacked together in little or no time, please don't take this too seriously.
 * Distributed under the WTFPL license (http://sam.zoy.org/wtfpl/COPYING),
 * which basically means you can do whatever you want with this.
 * 
 * Revision history:
 *  1.0     - Initial version
 *  1.1     - Added support for AuthKeys
 *  1.2     - Added support for non-.COM gTLDs and ccTLDs
 *          - AuthKeys may now include dashes (Jakub Bogusz, http://qboosh.pl/)
 *          - Existing files will not be overwritten, instead a "(2)"-type
 *            suffix will be added
 */

using System;
using System.IO;
using System.Net;
using System.Reflection;
using System.Text.RegularExpressions;
using System.Xml;

namespace PicasaWebDownloader
{
	/// <summary>
	/// Contains all relevant code for this project.
	/// </summary>
	public class Program
	{
	    /// <summary>
	    /// A pattern which matches gTLDs and ccTLDs.
	    /// Generated by <see cref="GenerateTLDRegex"/>.
	    /// </summary>
	    private const string TopLevelDomainPattern =
	        "AC|AD|AE|AERO|AF|AG|AI|AL|AM|AN|AO|AQ|AR|ARPA|AS|ASIA|AT|AU|AW|AX|AZ|BA" +
	        "|BB|BD|BE|BF|BG|BH|BI|BIZ|BJ|BM|BN|BO|BR|BS|BT|BV|BW|BY|BZ|CA|CAT|CC|CD" +
	        "|CF|CG|CH|CI|CK|CL|CM|CN|CO|COM|COOP|CR|CU|CV|CX|CY|CZ|DE|DJ|DK|DM|DO" +
	        "|DZ|EC|EDU|EE|EG|ER|ES|ET|EU|FI|FJ|FK|FM|FO|FR|GA|GB|GD|GE|GF|GG|GH|GI" +
	        "|GL|GM|GN|GOV|GP|GQ|GR|GS|GT|GU|GW|GY|HK|HM|HN|HR|HT|HU|ID|IE|IL|IM|IN" +
	        "|INFO|INT|IO|IQ|IR|IS|IT|JE|JM|JO|JOBS|JP|KE|KG|KH|KI|KM|KN|KP|KR|KW|KY" +
	        "|KZ|LA|LB|LC|LI|LK|LR|LS|LT|LU|LV|LY|MA|MC|MD|ME|MG|MH|MIL|MK|ML|MM|MN" +
	        "|MO|MOBI|MP|MQ|MR|MS|MT|MU|MUSEUM|MV|MW|MX|MY|MZ|NA|NAME|NC|NE|NET|NF" +
	        "|NG|NI|NL|NO|NP|NR|NU|NZ|OM|ORG|PA|PE|PF|PG|PH|PK|PL|PM|PN|PR|PRO|PS|PT" +
	        "|PW|PY|QA|RE|RO|RS|RU|RW|SA|SB|SC|SD|SE|SG|SH|SI|SJ|SK|SL|SM|SN|SO|SR" +
	        "|ST|SU|SV|SY|SZ|TC|TD|TEL|TF|TG|TH|TJ|TK|TL|TM|TN|TO|TP|TR|TRAVEL|TT|TV" +
	        "|TW|TZ|UA|UG|UK|US|UY|UZ|VA|VC|VE|VG|VI|VN|VU|WF|WS|YE|YT|YU|ZA|ZM|ZW";

        /// <summary>
        /// A regular expression used for matching PicasaWeb URLs, as wella s breaking them down into
        /// the parts necessary to generate the API URL.
        /// </summary>
        static readonly Regex PicasaWebURLPattern = new Regex(
                @"^(http://picasaweb\.google\.(?<tld>" + TopLevelDomainPattern + ")/" +
                @"((?<username>\w(\w|\.)+)/(?<album>\w(\w|\.)+))" +
                @"|(data/feed/base/user/(?<username>\w(\w|\.)+)/album(?<qual>id)/(?<album>\d+)))" +
                @"(\?.*(?<authkey>authkey\=(\w|-)+))?",
                RegexOptions.IgnoreCase | RegexOptions.ExplicitCapture | RegexOptions.IgnorePatternWhitespace |
                RegexOptions.Compiled );

        /// <summary>
        /// A replacement pattern, used to convert matches on <see cref="PicasaWebURLPattern"/> into their
        /// corresponding Google data API URLs.
        /// </summary>
        const string GoogleAPIReplacementPattern =
            "http://picasaweb.google.${tld}/data/feed/api/user/${username}/album${qual}/${album}?kind=photo&${authkey}";

	    /// <summary>
	    /// A regular expression used to match counter suffixes ("(2)", "(3)" etc.) in filenames.
	    /// </summary>
	    private static readonly Regex NumberedFilePattern =
	        new Regex( @"(?<prefix>.*)\((?<number>\d+)\)(?<ext>\.\w*)?$", RegexOptions.Compiled );

		/// <summary>
		/// Canonizes a PicasaWeb album web/RSS feed URL into its Google API equivalent.
		/// </summary>
		/// <param name="source">Source URL to be canonized</param>
		/// <param name="targetUri">Canonized URI instance</param>
		/// <returns><c>True</c> if the source URL was properly recognized, <c>false</c> otherwise</returns>
		/// <remarks>Currently only handles album and album feed URLs; makes many assumptions (including the
		/// PicasaWeb domain).</remarks>
		static bool canonizePicasaUrl( string source, out Uri targetUri )
		{
		    Match m = PicasaWebURLPattern.Match( source );
            targetUri = m.Success ? new Uri( m.Result( GoogleAPIReplacementPattern ).TrimEnd( '&' ) ) : null;
		    return m.Success;
		}
		
		/// <summary>
		/// Opens and downloads an image feed.
		/// </summary>
		/// <param name="feedUri">Feed URI</param>
		/// <param name="feed">Creates an XML document containing the downloaded feed</param>
		/// <returns><c>True</c> if the feed was successfully downloaded, <c>false</c> otherwise</returns>
		static bool openFeed( Uri feedUri, out XmlDocument feed )
		{
			feed = null;
			
			Console.Write( "Opening feed..." );
			try
			{
				WebRequest req = WebRequest.Create( feedUri );
				using ( WebResponse resp = req.GetResponse() )
				using ( MemoryStream ms = new MemoryStream() )
				{
					// Download feed
					download( resp, ms, delegate( long offset, long len ) {
					         	Console.Write( "\rDownloading feed ({0}KB)", offset / 1024 ); 
					         } );
					Console.WriteLine( "\rProcessing feed          " );
					
					// Load feed into an XmlDocument instance
					ms.Seek( 0, SeekOrigin.Begin );
					try
					{
						feed = new XmlDocument();
						feed.Load( ms );
						return true;
					}
					catch ( XmlException e )
					{
						Console.WriteLine( "Invalid feed: {0}", e.Message );
						return false;
					}
				}
			}
			catch ( Exception e )
			{
				Console.WriteLine( "\rCan't open feed: {0}", e.Message );
				return false;
			}
		}
		
		/// <summary>
		/// Processes the PicasaWeb API ATOM feed and creates a list of image entries.
		/// </summary>
		/// <param name="feed">XML document containing the PicasaWeb API ATOM feed</param>
		/// <returns>A list of image entries for the specified feed</returns>
		static string[] processEntries( XmlDocument feed )
		{
			// Create a namespace manager (otherwise we can't do squat)
			XmlNamespaceManager xnsm = new XmlNamespaceManager( feed.NameTable );
            xnsm.AddNamespace( "atom", "http://www.w3.org/2005/Atom" );

            // Convoluted XPath expression for extracting the source URLs for all images in the feed
            XmlNodeList entries =
                feed.SelectNodes( "atom:feed/atom:entry[atom:category/@scheme=\"http://schemas.google.com/g/2005#kind\" " +
                                                       "and atom:category/@term=\"http://schemas.google.com/photos/2007#photo\"]" +
                                  "/atom:content/@src", xnsm );
            
            // Create and return the result set
            string[] resultSet = new string[ entries.Count ];
            int i = 0;
            foreach ( XmlNode node in entries )
            	resultSet[ i++ ] = node.Value;
            return resultSet;
		}
		
		/// <summary>
		/// Delegate describing an update request in the download manager
		/// </summary>
		/// <param name="offset">Current position in the downloaded stream</param>
		/// <param name="contentLength">Content length for the downloaded stream</param>
		/// <remarks>Content length is directly derived from the <see cref="T:System.Net.WebResponse"/> and
		/// inherits its quirks.</remarks>
		delegate void DownloadUpdateDelegate( long offset, long contentLength );

		/// <summary>
		/// Streams the entire specified web response, calling a user-defined update function in 16KB increments.
		/// </summary>
		/// <param name="resp">Web response from which to download the file</param>
		/// <param name="target">Target stream</param>
		/// <param name="update">User-defined update delegate for progress report in the UI</param>		
		static void download( WebResponse resp, Stream target, DownloadUpdateDelegate update )
		{
			download( resp, target, 16384, update );
		}
		
		/// <summary>
		/// Streams the entire specified web response, calling a user-defined update function when required.
		/// </summary>
		/// <param name="resp">Web response from which to download the file</param>
		/// <param name="target">Target stream</param>
		/// <param name="updateInterval">Block size (in bytes) between calls to <paramref name="update"/></param>
		/// <param name="update">User-defined update delegate for progress report in the UI</param>		
		static void download( WebResponse resp, Stream target, int updateInterval, DownloadUpdateDelegate update )
		{
			long offset = 0;
			byte[] buffer = new byte[ updateInterval ];
			using ( Stream source = resp.GetResponseStream() )
			{
				int count;
				do
				{
					count = source.Read( buffer, 0, buffer.Length );
					target.Write( buffer, 0, count );
					update( offset += count, resp.ContentLength );
				} while ( count > 0 );
			}
		}
		
		/// <summary>
		/// Downloads a single feed entry.
		/// </summary>
		/// <param name="prefix">Prefix to be shown before the download progress update</param>
		/// <param name="targetFolder">Target folder in which the downloaded file is to be saved</param>
		/// <param name="url">Download URL</param>
		static void downloadFile( string prefix, string targetFolder, string url )
		{
			// Parse URI
			Uri uri;
        	if ( !Uri.TryCreate( url, UriKind.Absolute, out uri ) )
        		Console.WriteLine( "Invalid image URL: " + url );
        	else
        	{
        		string filename = addSuffixToFile( Uri.UnescapeDataString( uri.Segments[ uri.Segments.GetUpperBound( 0 ) ] ) );
            	string msg      = "\r" + prefix + "{0,5}/{1,5}kb\t{2}";
            	string final    = "\r" + prefix + "       Done  ";
            	Console.Write( msg, '?', '?', filename );
            	
            	try
            	{
            		// Create web request and target file stream
            		HttpWebRequest req = (HttpWebRequest) WebRequest.Create( uri );
            		using ( HttpWebResponse resp = (HttpWebResponse) req.GetResponse() )
            		using ( Stream target = File.Create( Path.Combine( targetFolder, filename ) ) )
            		{
						download( resp, target, delegate( long offset, long len ) {
            			         	Console.Write( msg, offset / 1024, len > 0 ? len / 1024 : '?', filename );
            			         } );
						Console.WriteLine( final );
            		}
            	}
            	catch ( Exception e )
            	{
            		Console.WriteLine( "\tError: " + e.Message );
            	}
        	}
        	
		}

	    /// <summary>
	    /// Locates the specified file. If it already exists, a counter suffix (e.g. "(2)", "(3)" etc.) is added.
	    /// </summary>
	    /// <param name="filename">The filename for which to add the counter suffix.</param>
	    /// <returns>The target filename, including the counter suffix (if applicable).</returns>
	    private static string addSuffixToFile( string filename )
	    {
	        if ( !File.Exists( filename ) )
	            return filename;

	        Match m = NumberedFilePattern.Match( filename );
	        if ( m.Success ) 
            {
                int count = int.Parse( m.Groups[ "number" ].Value ) + 1;
	            return addSuffixToFile( m.Result( "${prefix}(" + count + ")${ext}" ) );
            }
	        else
	            return addSuffixToFile(
	                Path.Combine( Path.GetDirectoryName( filename ),
	                              Path.GetFileNameWithoutExtension( filename ) + " (2)" + Path.GetExtension( filename ) ) );
	    }

	    /// <summary>
		/// Driver script for this project.
		/// </summary>
		/// <param name="args">Command line arguments</param>
        public static void Main( string[] args )
        {
	        AssemblyName program = Assembly.GetExecutingAssembly().GetName();
       		Console.WriteLine( "PicasaWeb Downloader v" + program.Version + " by Tomer Gabel, 2007-2008" );
       		Console.WriteLine();

       		// Wrong arguments? Show usage
       		if ( ( args.Length < 1 ) || ( args.Length > 2 ) )
        	{
        		Console.WriteLine( "Usage:" );
        		Console.WriteLine( "\t" + program.Name + " <url> [target]");
        		Console.WriteLine();
        		Console.WriteLine( "\turl\tURL for the album or its RSS feed" );
        		Console.WriteLine( "\ttarget\tTarget folder for download (e.g. \"C:\\My Pictures\")" );
        		Console.WriteLine( "\t\tIf no target is specified, the current folder is used." );
        		return;
        	}
        	
       		// Attempt to parse and canonize the specified URL
        	Uri feedUri;
        	if ( !canonizePicasaUrl( args[ 0 ], out feedUri ) )
        	{
        		Console.WriteLine( "Unrecognized PicasaWeb URL " + args[ 0 ] );
        		return;
        	}
        	
        	// Set up and, if necessary, create target folder
        	string targetFolder = Directory.GetCurrentDirectory();
        	if ( args.Length > 1 )
        	{
        		targetFolder = args[ 1 ];
        		try 
        		{
        			Directory.CreateDirectory( targetFolder ); 
        		}
        		catch ( Exception e ) 
        		{
        			Console.WriteLine( "Can't create target folder " + targetFolder + ": " + e.Message );
        			return;
        		}
        	}
        	
        	// Open the source feed
        	XmlDocument feed;
        	if ( !openFeed( feedUri, out feed ) )
        	{
        		Console.WriteLine( "Can't open feed " + feedUri.AbsoluteUri );
        		return;
        	}
        	
        	// Process feed, iterate and process entries
        	string[] entries = processEntries( feed );
        	if ( entries.Length > 0 )
        	{
        		int i = 0;
        		foreach ( string url in entries )
        			downloadFile( string.Format( "#{0,3}/{1}\t", ++i, entries.Length ), targetFolder, url );
        	
        		Console.WriteLine();
        		Console.WriteLine( "All done." );
        	}
        	else
        		Console.WriteLine( "No images found in feed!" );
        }
	}
}
